//! # Tasoller CFW1.0
//! USB device: 1CCF:2333
//!
//! USB interface: 0
//!
//! ## Protocol information
//!
//! ### IN Interrupt (0x84)
//! It is more useful to think of the input as a sequence of bits,
//! going from the least significant bit to the most significant.
//!
//! ```
//! let bits = input
//!     .iter()
//!     .flat_map(|x| (0..8).map(move |i| (x >> i) & 1));
//! ```
//!
//! - Content length: 11 bytes (88 bits)
//! - Bits 0..26: ???
//! - Bit 26: FN2 (1 = pressed)
//! - Bit 27: FN1 (1 = pressed)
//! - Bit 28: IR2
//! - Bit 29: IR1
//! - Bit 30: IR4
//! - Bit 31: IR3
//! - Bit 32: IR6
//! - Bit 33: IR5
//! - Bits 34..66: touch sensor 1-32 (1 = held, top -> bottom then left -> right)
//! - Bits 66..88: ???
//!
//! ### OUT Bulk (0x03)
//! - Content length: 240 bytes
//! - Bytes 0..3: magic bytes [0x42, 0x4C, 0x00]
//! - Bytes 3..96: Slider LED (BGR order, right -> left)
//! - Bytes 96..168: Left tower LEDs (GRB)
//! - Bytes 168..240: Right tower LEDs (GRB)

#[cfg(any(chuni, chusanapp))]
use anyhow::Result;
#[cfg(any(chuni, chusanapp))]
use rusb::{DeviceHandle, UsbContext};

use super::ReadType;
#[cfg(any(chuni, chusanapp))]
use crate::TIMEOUT;

pub const DEVICE_VID: u16 = 0x1CCF;
pub const DEVICE_PID: u16 = 0x2333;
pub const READ_TYPE: ReadType = ReadType::Interrupt;
pub const READ_ENDPOINT: u8 = 0x84;
pub const INPUT_MEMORY_SIZE: usize = 11;
pub const OUTPUT_MEMORY_SIZE: usize = 240;

#[cfg(any(chuni, amdaemon))]
#[inline(always)]
pub fn jvs_poll(input: &[u8]) -> (u8, u8) {
    // FN1 | FN2
    let opbtn = input[3] >> 3 & 1 
        | input[3] >> 1 & 2;
    
    // IR1-6 in order
    let beams = input[3] >> 5 & 1
        | input[3] >> 3 & 2
        | input[3] >> 5 & 4
        | input[3] >> 3 & 8
        | input[4] << 3 & 16
        | input[4] << 5 & 32;

    (opbtn, beams)
}

#[cfg(any(chuni, amdaemon))]
#[inline(always)]
pub fn is_coin_button_pressed(_input: &[u8]) -> bool {
    // Tasoller does not have a coin button
    false
}

#[cfg(any(chuni, chusanapp))]
#[inline(always)]
pub fn read_pressure_data(input: &[u8]) -> [u8; 32] {
    let bits = input
        .iter()
        .skip(4) // Skip first 4 bytes
        .flat_map(|x| (0..8).map(move |i| (x >> i) & 1))
        .skip(2) // Skip next 2 bits
        .take(32)
        .enumerate();
    let mut pressure = [0u8; 32];

    for (i, bit) in bits {
        pressure[if i % 2 == 0 { 30 - i } else { 32 - i }] = bit * 255;
    }

    pressure
}

#[cfg(any(chuni, chusanapp))]
#[inline(always)]
pub fn init_output_buffer(output: &mut [u8]) {
    output[0] = 0x42;
    output[1] = 0x4C;
    output[2] = 0;
}

#[cfg(any(chuni, chusanapp))]
#[inline(always)]
pub fn set_slider_leds<T: UsbContext>(
    device: &DeviceHandle<T>,
    output: &mut [u8],
    rgb: &[u8],
) -> Result<()> {
    for (buf_chunk, state_chunk) in output[3..96].chunks_mut(3).take(31).zip(rgb.chunks(3)) {
        buf_chunk[0] = state_chunk[2];
        buf_chunk[1] = state_chunk[1];
        buf_chunk[2] = state_chunk[0];
    }

    device.write_bulk(0x03, output, TIMEOUT)?;

    Ok(())
}

#[cfg(any(chuni, chusanapp))]
#[inline(always)]
pub fn set_led_colors<T: UsbContext>(
    device: &DeviceHandle<T>,
    output: &mut [u8],
    board: u8,
    rgb: &[u8],
) -> Result<()> {
    if board != 0 && board != 1 {
        return Ok(());
    }

    let air_rgb = if board == 0 { 0x96 } else { 0xB4 };
    let output_start = if board == 0 { 96 } else { 168 };

    for buf_chunk in output[output_start..output_start + 72]
        .chunks_mut(3)
        .take(24)
    {
        buf_chunk[0] = rgb[air_rgb + 1];
        buf_chunk[1] = rgb[air_rgb];
        buf_chunk[2] = rgb[air_rgb + 2];
    }

    device.write_bulk(0x03, output, TIMEOUT)?;

    Ok(())
}
