//! # Tasoller CFW2.0
//! USB device: 1CCF:2333
//!
//! USB interface: 0
//!
//! ## Protocol information
//!
//! ### IN Interrupt (0x84)
//! - Content length: 36 bytes
//! - Bytes 0..3: Magic bytes [0x68, 0x66, 0x84]
//! - Byte 3
//!   - Bit 0..6: ir 1-6 (1 = blocked)
//!   - Bit 6: FN1 (1 = pressed)
//!   - Bit 7: FN2 (1 = pressed)
//! - Bytes 4..36: Touch sensor pressure (starts from top left,
//! top to bottom then left to right)
//!
//! ### OUT Bulk (0x03)
//! - Content length: 240 bytes
//! - Bytes 0..3: magic bytes [0x42, 0x4C, 0x00]
//! - Bytes 3..96: Slider LED (GRB order, right -> left)
//! - Bytes 96..168: Left tower LEDs (GRB)
//! - Bytes 168..240: Right tower LEDs (GRB)

#[cfg(any(chuni, chusanapp))]
use anyhow::Result;
#[cfg(any(chuni, chusanapp))]
use rusb::{DeviceHandle, UsbContext};

use super::ReadType;
#[cfg(any(chuni, chusanapp))]
use crate::TIMEOUT;

pub const DEVICE_VID: u16 = 0x1CCF;
pub const DEVICE_PID: u16 = 0x2333;
pub const READ_TYPE: ReadType = ReadType::Interrupt;
pub const READ_ENDPOINT: u8 = 0x84;
pub const INPUT_MEMORY_SIZE: usize = 36;
pub const OUTPUT_MEMORY_SIZE: usize = 240;

#[cfg(any(chuni, amdaemon))]
#[inline(always)]
pub fn jvs_poll(input: &[u8]) -> (u8, u8) {
    let opbtn = input[3] >> 6;
    let beams = input[3] & 63; // mask for last 6 bits

    (opbtn, beams)
}

#[cfg(any(chuni, amdaemon))]
#[inline(always)]
pub fn is_coin_button_pressed(_input: &[u8]) -> bool {
    // Tasoller does not have a coin button
    false
}

#[cfg(any(chuni, chusanapp))]
#[inline(always)]
pub fn read_pressure_data(input: &[u8]) -> [u8; 32] {
    let mut pressure = [0u8; 32];

    for (i, p) in input.iter().skip(4).take(32).enumerate() {
        pressure[if i % 2 == 0 { 30 - i } else { 32 - i }] = *p;
    }

    pressure
}

#[cfg(any(chuni, chusanapp))]
#[inline(always)]
pub fn init_output_buffer(output: &mut [u8]) {
    output[0] = 0x42;
    output[1] = 0x4C;
    output[2] = 0;
}

#[cfg(any(chuni, chusanapp))]
#[inline(always)]
pub fn set_slider_leds<T: UsbContext>(
    device: &DeviceHandle<T>,
    output: &mut [u8],
    rgb: &[u8],
) -> Result<()> {
    for (buf_chunk, state_chunk) in output[3..96].chunks_mut(3).take(31).zip(rgb.chunks(3)) {
        buf_chunk[0] = state_chunk[1];
        buf_chunk[1] = state_chunk[0];
        buf_chunk[2] = state_chunk[2];
    }

    device.write_bulk(0x03, output, TIMEOUT)?;

    Ok(())
}

#[cfg(any(chuni, chusanapp))]
#[inline(always)]
pub fn set_led_colors<T: UsbContext>(
    device: &DeviceHandle<T>,
    output: &mut [u8],
    board: u8,
    rgb: &[u8],
) -> Result<()> {
    if board != 0 && board != 1 {
        return Ok(());
    }

    let air_rgb = if board == 0 { 0x96 } else { 0xB4 };
    let output_start = if board == 0 { 96 } else { 168 };

    for buf_chunk in output[output_start..output_start + 72]
        .chunks_mut(3)
        .take(24)
    {
        buf_chunk[0] = rgb[air_rgb + 1];
        buf_chunk[1] = rgb[air_rgb];
        buf_chunk[2] = rgb[air_rgb + 2];
    }

    device.write_bulk(0x03, output, TIMEOUT)?;

    Ok(())
}
